unit pwr;

//     POWER8


// ***************************************************************************************************
// 
// ***************************************************************************************************
interface
Uses Classes, Windows, SysUtils, Dialogs;


//  
type
	TPWRCommParam = packed record
		BaudRate, Parity 	: Byte;
	End;

	TPWRParam = packed record
		Flags				: Byte;
		SoftTime			: Byte;
		PassWord			: Array[0..7] of Byte;
	End;

	TPWRState = packed record
		Flags				: Byte;
		Leds				: Byte;
	End;


Const
//   
	FLG_CLK_SOUND			= 1;
	FLG_LOCALMODE			= 2;
	FLG_SOFTCONTROL			= 4;
	FLG_PERSIST				= 8;

//    
	FLG_STATE_LOCALMODE		= 1;


//  
	REQ_BLOCK				= $04;
	REQ_RESET				= $05;


	REQ_GET_COMMPARAM		= $1A;
	REQ_SET_COMMPARAM		= $1B;

	REQ_GET_PWRPARAM		= $40;
	REQ_SET_PWRPARAM		= $41;
	REQ_ON_CHANELS			= $42;
	REQ_OFF_CHANELS			= $43;
	REQ_ON_CHANEL			= $44;
	REQ_OFF_CHANEL			= $45;
	REQ_SET_CHANELS			= $46;
	REQ_GET_CHANELS			= $47;
	REQ_GET_STATE			= $48;
	REQ_SET_STATE			= $49;


//  
	pwrERROR_NO_ERROR			= 0;
	pwrERROR_INV_UNSUPPORTED	= 150;
	pwrERROR_INV_PARAM			= 151;
	pwrERROR_INV_FORMAT			= 152;


//  
Procedure	ClosePort;
Procedure	DetectPorts(List : TStrings);
Function	OpenPort(Port : String; Baud : Integer; Parity : Integer) : Boolean;
Function	pwrReadChanels : Boolean;
Function	pwrGetChanelState(No : Integer) : Boolean;
Procedure	pwrBeginChanelUpdate;
Procedure	pwrSetChanelState(no : Integer; state : Boolean);
Function	pwrEndChanelUpdate : Boolean;
function 	pwrGetCommParam(var param : TPWRCommParam) : Boolean;
function 	pwrSetCommParam(param : TPWRCommParam) : Boolean;
function 	pwrGetParam(var param : TPWRParam) : Boolean;
function 	pwrSetParam(param : TPWRParam) : Boolean;
function	pwrBlock : Boolean;
function	pwrReset : Boolean;
function	pwrGetState(var State : TPWRState) : Boolean;
function	pwrSetState(State : TPWRState) : Boolean;

var
	ChanelCount		: Integer;


// ***************************************************************************************************
// 
// ***************************************************************************************************
implementation

Var
	Chanels 	: Array[0..63] of Boolean;						//  
	ChanelsOn 	: Array[0..63] of Boolean;						//    
	ChanelsOff 	: Array[0..63] of Boolean;						//    
	hPort		: THandle = INVALID_HANDLE_VALUE;				//    


// ***************************************************************************************************
//    
// ***************************************************************************************************
Procedure DetectPorts(List : TStrings);
Var
	i 		: Integer;
	cconf	: TCommConfig;
	n		: DWord;
Begin
	List.Clear;
	for i := 1 to 64 do Begin
		n := sizeof(TCommConfig);
		if GetDefaultCommConfig(PChar('COM'+IntToStr(i)), cconf, n) Then List.Add('COM'+IntToStr(i));
	End;
End;




// ***************************************************************************************************
//  
// ***************************************************************************************************
Procedure ClosePort;
Begin
	if hPort <> INVALID_HANDLE_VALUE Then CloseHandle(hPort);
	hPort := INVALID_HANDLE_VALUE;
End;





// ***************************************************************************************************
//  
// ***************************************************************************************************
Function OpenPort(Port : String; Baud : Integer; Parity : Integer) : Boolean;
Var
	TimeOut : TCommTimeOUts;
	dcb     : TDCB;
Begin
	Result := False;

	//      
	ClosePort();

	//  
	hPort := CreateFile(Pchar(Port),
				GENERIC_READ or GENERIC_WRITE,
				0,											// comm devices must be opened w/exclusive-access
				nil,										// no security attributes
				OPEN_EXISTING,								// comm devices must use OPEN_EXISTING
				0,											// no overlapped I/O
                0											// hTemplate must be NULL for comm devices
			);

    if hPort = INVALID_HANDLE_VALUE Then Begin
		ShowMessage('    (CreateFile)');
		Exit;
	End;


	//  
	if not GetCommTimeOuts(hPort, TimeOut) Then Begin
		ClosePort();
		ShowMessage('    (GetCommTimeOuts)');
		Exit;
	End;

	//  
	TimeOut.ReadIntervalTimeout			:= 5+(1000 div Baud);
	TimeOut.ReadTotalTimeoutMultiplier	:= (1000 div Baud) + 1;
	TimeOut.ReadTotalTimeoutConstant	:= 400;

	if not SetCommTimeOuts(hPort, TimeOut) Then Begin
		ClosePort();
		ShowMessage('    (SetCommTimeOuts)');
		Exit;
	End;


	if not GetCommState(hPort, dcb) Then Begin
		ClosePort();
		ShowMessage('    (GetCommState)');
		Exit;
	End;

	//   
	case parity of
		0 : dcb.Parity := NOPARITY;
		1 : dcb.Parity := MARKPARITY;
		2 : dcb.Parity := SPACEPARITY;
		3 : dcb.Parity := EVENPARITY;
		4 : dcb.Parity := ODDPARITY;
		else dcb.Parity := NOPARITY;
	end;
	dcb.ByteSize	:= 8;
	dcb.StopBits	:= ONESTOPBIT;
	dcb.Flags		:= dcb.Flags and (not $3030);			//   
	dcb.Flags		:= dcb.Flags or $1010;					// RTS/CTS   
	dcb.BaudRate := Baud;

	if not SetCommState(hPort,dcb) then Begin
		ClosePort();
		ShowMessage('    (SetCommTimeOuts)');
		Exit;
	End;

	Result := True;
End;






// ***************************************************************************************************
//   CRC8    
// ***************************************************************************************************
function CalcCrc_8(ptr : PByte; count : Integer) : Byte;
Var
	i : Integer;
	b : Byte;
begin
	Result := $FF;
	While Count <> 0 do Begin
		b := ptr^; inc(ptr);

		for i := 0 to 7 do Begin
			if (Result xor b) and 1 = 1 then Begin
				Result := Result shr 1;
				Result := Result xor $8c;
			End else Result := Result shr 1;
			b := b shr 1;
		End;

		dec(count);
	End;
End;





// ***************************************************************************************************
//   ,      
// ***************************************************************************************************
function SendReq(Var ReqCode : Byte; Var Data : String) : Boolean;
Var
	dat : String;
	n : Dword;
Begin
	Result := False;

	if hPort = INVALID_HANDLE_VALUE Then Exit;					//   

	//     
	dat := Char(ReqCode) + Data;
	//   
	dat := Char(Length(Dat)+2) + Dat;
	//     
	dat := Char($E2) + dat + Char(CalcCrc_8(@dat[1], Length(Dat)));

    //  
	if not PurgeComm(hPort, PURGE_TXCLEAR or PURGE_RXCLEAR) Then Exit;
	if not WriteFile(hPort, dat[1], Length(dat), n, nil) Then Exit;
	SetLength(dat, 100);
    //      -  
	//   (   )    
	//        
	if not ReadFile(hPort, dat[1], 100, n, nil) Then Exit;

	//   

	//       4-  (  +   +
    //      +  )
	if n < 4 Then Exit;
	//    
    if Byte(dat[1]) <> $E2 Then Exit;
	//         
	if Byte(dat[2]) <> n-1 Then Exit;
	//   
	if CalcCRC_8(@dat[2], n-1) <>0 Then Exit;


	//  ,  
	ReqCode	:= Byte(dat[3]);
	Data 	:= Copy(dat,4, n-4);

	Result := True;
End;




// ***************************************************************************************************
//     
// ***************************************************************************************************
Function pwrReadChanels : Boolean;
Var
	ReqCode : Byte;
	Data	: String;
	i		: Integer;
begin
	Result := False;

	ReqCode := REQ_GET_CHANELS;
	Data := '';

	if not SendReq(ReqCode, Data) Then Exit;
	if ReqCode <> pwrERROR_NO_ERROR Then Exit;

	//     
	FillChar(Chanels, SizeOf(Chanels), 0);
    ChanelCount := Length(Data) * 8;
	For i := 0 to ChanelCount-1 do
		Chanels[i] := (Byte(Data[1 + (i div 8)]) and (1 shl (i and 7))) <> 0;

	Result := True;
End;




// ***************************************************************************************************
//     
// ***************************************************************************************************
Function pwrGetChanelState(No : Integer) : Boolean;
Begin
	Result := Chanels[no];
End;



// ***************************************************************************************************
//    
// ***************************************************************************************************
Procedure pwrBeginChanelUpdate;
Begin
	//     
	FillChar(ChanelsOn, sizeof(ChanelsOn), 0);
	FillChar(ChanelsOff, sizeof(ChanelsOff), 0);
End;




// ***************************************************************************************************
//   
// ***************************************************************************************************
Procedure pwrSetChanelState(no : Integer; state : Boolean);
Begin
	ChanelsOn[no]	:= state;
	ChanelsOff[no]	:= not state;
End;







// ***************************************************************************************************
//  ,   
// ***************************************************************************************************
Function pwrEndChanelUpdate : Boolean;
var
	OnCount, OffCount,
	OnStart, OffStart, i	: Integer;
	OnMask, OffMask			: Array[0..7] of Byte;
	Data					: String;
	ReqCode					: Byte;
Begin
    Result := False;

	//     
	OnCount := 0; OnStart := -1; FillChar(OnMask, sizeof(OnMask),0);
	For i := 0 to ChanelCount-1 do Begin
		If ChanelsOn[i] Then Begin
			if OnStart = -1 Then OnStart := i;
			OnMask[(i-OnStart) div 8] := OnMask[(i-OnStart) div 8] or (1 shl ((i-OnStart) and 7));
			OnCount := i - OnStart + 1;
		End;
	End;


	//     
	OffCount := 0; OffStart := -1; FillChar(OffMask, sizeof(OffMask),0);
	For i := 0 to ChanelCount-1 do Begin
		If ChanelsOff[i] Then Begin
			if OffStart = -1 Then OffStart := i;
			OffMask[(i-OffStart) div 8] := OffMask[(i-OffStart) div 8] or (1 shl ((i-OffStart) and 7));
			OffCount := i - OffStart + 1;
		End;
	End;


	//   
    If OnCount<>0 Then Begin
		if OnCount = 1 Then Begin
			//   ,   
			ReqCode := REQ_ON_CHANEL;
			Data := Char(OnStart);
			if not SendReq(ReqCode, Data) Then Exit;
			if ReqCode <> pwrERROR_NO_ERROR Then Exit;
		End else Begin
			//   
			ReqCode := REQ_ON_CHANELS;
			//      
			i := (OnCount + 7) div 8;
			SetLength(Data, i);
			Move(OnMask, Data[1], i);
			Data := Char(OnStart)+Data;
			if not SendReq(ReqCode, Data) Then Exit;
			if ReqCode <> pwrERROR_NO_ERROR Then Exit;
		End;
	End;

	//   
    If OffCount<>0 Then Begin
		if OffCount = 1 Then Begin
			//   ,   
			ReqCode := REQ_OFF_CHANEL;
			Data := Char(OffStart);
			if not SendReq(ReqCode, Data) Then Exit;
			if ReqCode <> pwrERROR_NO_ERROR Then Exit;
		End else Begin
			//   
			ReqCode := REQ_OFF_CHANELS;
			//      
			i := (OffCount + 7) div 8;
			SetLength(Data, i);
			Move(OffMask, Data[1], i);
			Data := Char(OffStart)+Data;
			if not SendReq(ReqCode, Data) Then Exit;
			if ReqCode <> pwrERROR_NO_ERROR Then Exit;
		End;
	End;

	Result := True;
End;




// ***************************************************************************************************
//   
// ***************************************************************************************************
function pwrGetCommParam(var param : TPWRCommParam) : Boolean;
Var
	ReqCode : Byte;
	Data : String;
Begin
    Result := False;
	ReqCode := REQ_GET_COMMPARAM;
	Data := '';
	if not SendReq(ReqCode, Data) Then Exit;
	if ReqCode <> pwrERROR_NO_ERROR Then Exit;
	if Length(Data) <> sizeof(TPWRCommParam) then Exit;
	Move(Data[1], param, sizeof(TPWRCommParam));
	Result := True;
End;



// ***************************************************************************************************
//   
// ***************************************************************************************************
function pwrSetCommParam(param : TPWRCommParam) : Boolean;
Var
	ReqCode : Byte;
	Data : String;
Begin
    Result := False;
	ReqCode := REQ_SET_COMMPARAM;
	SetLength(Data, sizeof(TPWRCommParam));
	Move(param, Data[1], sizeof(TPWRCommParam));
	if not SendReq(ReqCode, Data) Then Exit;
	if ReqCode <> pwrERROR_NO_ERROR Then Exit;
	Result := True;
End;




// ***************************************************************************************************
//  PWR 
// ***************************************************************************************************
function pwrGetParam(var param : TPWRParam) : Boolean;
Var
	ReqCode : Byte;
	Data : String;
Begin
    Result := False;
	ReqCode := REQ_GET_PWRPARAM;
	Data := '';
	if not SendReq(ReqCode, Data) Then Exit;
	if ReqCode <> pwrERROR_NO_ERROR Then Exit;
	if Length(Data) <> sizeof(TPWRParam) then Exit;
	Move(Data[1], param, sizeof(TPWRParam));
	Result := True;
End;



// ***************************************************************************************************
//  PWR 
// ***************************************************************************************************
function pwrSetParam(param : TPWRParam) : Boolean;
Var
	ReqCode : Byte;
	Data : String;
Begin
    Result := False;
	ReqCode := REQ_SET_PWRPARAM;
	SetLength(Data, sizeof(TPWRParam));
	Move(param, Data[1], sizeof(TPWRParam));
	if not SendReq(ReqCode, Data) Then Exit;
	if ReqCode <> pwrERROR_NO_ERROR Then Exit;
	Result := True;
End;




// ***************************************************************************************************
//  
// ***************************************************************************************************
function	pwrBlock : Boolean;
Var
	ReqCode : Byte;
	Data : String;
Begin
    Result := False;
	ReqCode := REQ_BLOCK; Data := '';
	if not SendReq(ReqCode, Data) Then Exit;
	if ReqCode <> pwrERROR_NO_ERROR Then Exit;
	Result := True;
End;




// ***************************************************************************************************
//  ( )
// ***************************************************************************************************
function	pwrReset : Boolean;
Var
	ReqCode : Byte;
	Data : String;
Begin
    Result := False;
	ReqCode := REQ_RESET; Data := '';
	if not SendReq(ReqCode, Data) Then Exit;
	if ReqCode <> pwrERROR_NO_ERROR Then Exit;
	Result := True;
End;



// ***************************************************************************************************
//    
// ***************************************************************************************************
function pwrGetState(var State : TPWRState) : Boolean;
Var
	ReqCode : Byte;
	Data : String;
Begin
    Result := False;
	ReqCode := REQ_GET_STATE;
	Data := '';
	if not SendReq(ReqCode, Data) Then Exit;
	if ReqCode <> pwrERROR_NO_ERROR Then Exit;
	if Length(Data) <> sizeof(TPWRState) then Exit;
	Move(Data[1], State, sizeof(TPWRState));
	Result := True;
End;




// ***************************************************************************************************
//   
// ***************************************************************************************************
function pwrSetState(State : TPWRState) : Boolean;
Var
	ReqCode : Byte;
	Data : String;
Begin
    Result := False;
	ReqCode := REQ_SET_STATE;
	SetLength(Data, sizeof(TPWRState));
	Move(State, Data[1], sizeof(TPWRState));
	if not SendReq(ReqCode, Data) Then Exit;
	if ReqCode <> pwrERROR_NO_ERROR Then Exit;
	Result := True;
End;



end.
